/* -*- tab-width: 4 -*- */

#include <ctype.h>

#include "global.h"
#include "elkkernel.h"

ELKObject Arg_True;

static ELKObject V_Error_Handler, V_Top_Level_Control_Point;

void Err_Handler (ELKObject sym, ELKObject fmt, int argc, ELKObject *argv);

/* Error_Tag should be static and users should only use the functions
 * Set_Error_Tag() and Get_Error_Tag().  However, in older versions
 * the variable was manipulated directly, therefore it will remain global
 * for some time for backwards compatibility.
 */
const CHAR *Error_Tag;

CHAR *appname;

void Init_Error (void) {
	Arg_True = Cons (Trueob, Null);
	Global_GC_Link (Arg_True);
	Define_Variable (&V_Error_Handler, x_("error-handler"), Null);
	Define_Variable (&V_Top_Level_Control_Point, x_("top-level-control-point"),
		Null);
}

const CHAR *Get_Error_Tag (void) {
	return Error_Tag;
}

void Set_Error_Tag (const CHAR *tag) {
	Error_Tag = tag;
}

void Set_App_Name (CHAR *name) {
	appname = name;
}

void Fatal_Error (CHAR *fmt, ...) {
	va_list args;
	CHAR line[500];

	Disable_Interrupts;
	var_start(args, fmt);
	(void)evsprintf(line, fmt, args);
	va_end (args);
	error(line);
}

void Panic (const CHAR *msg) {
	Disable_Interrupts;
	if (appname)
		error(x_("%s: panic: %s"), appname, msg);
	else
		error(x_("Panic: %s"), msg);
}

void Uncatchable_Error (CHAR *errmsg) {
	Disable_Interrupts;
	Reset_IO (0);
	/*
	 * The message can be sent to stdout, as Reset_IO() resets the
	 * current output port back to the Standard_Output_Port:
	 */
	ttyputerr(x_("%s"), errmsg);
	Reset ();
}

void Primitive_Error (CHAR *fmt, ...) {
	va_list args;
	register CHAR *p;
	register int i, n;
	ELKObject msg, sym, argv[10];
	GC_Node; GCNODE gcv;

	var_start (args, fmt);
	for (n = 0, p = fmt; *p; p++)
		if (*p == '~' && p[1] != '~' && p[1] != '%'
				&& p[1] != 'E' && p[1] != 'e')
			n++;
	if (n > 10)
		Panic (x_("Primitive_Error args"));
	for (i = 0; i < n; i++)
		argv[i] = va_arg (args, ELKObject);
	sym = Null;
	GC_Link (sym);
	gcv.gclen = 1 + i; gcv.gcobj = argv; gcv.next = &gc1; GC_List = &gcv;
	sym = Intern (Error_Tag);
	msg = Make_String (fmt, p - fmt);
	Err_Handler (sym, msg, i, argv);
	/*NOTREACHED*/
}

ELKObject P_Error (int argc, ELKObject *argv) {
	Check_Type (argv[1], T_String);
	Err_Handler (argv[0], argv[1], argc-2, argv+2);
	/*NOTREACHED*/
	return(Null);
}

void Err_Handler (ELKObject sym, ELKObject fmt, int argc, ELKObject *argv) {
	ELKObject fun, args, a[1];
	GC_Node3;

	Reset_IO (0);
	args = Null;
	GC_Link3 (args, sym, fmt);
	args = P_List (argc, argv);
	args = Cons (fmt, args);
	args = Cons (sym, args);
	fun = Var_Get (V_Error_Handler);
	if (TYPE(fun) == T_Compound)
		(void)Funcall (fun, args, 0);
	a[0] = sym;
	Format (Curr_Output_Port, x_("~s: "), 4, 1, a);
	Format (Curr_Output_Port, STRING(fmt)->data, STRING(fmt)->size,
		argc, argv);
	(void)P_Newline (0, (ELKObject *)0);
	GC_Unlink;
	Reset ();
	/*NOTREACHED*/
}

void Reset (void) {
	ELKObject cp;

	cp = Var_Get (V_Top_Level_Control_Point);
	if (TYPE(cp) == T_Control_Point)
		(void)Funcall_Control_Point (cp, Arg_True, 0);
	(void)fflush (stdout);
	error(x_("LISP Error"));
}

ELKObject P_Reset (void) {
	Reset_IO (0);
	Reset ();
	/*NOTREACHED*/
	return(Null);
}

void Range_Error (ELKObject i) {
	Primitive_Error (x_("argument out of range: ~s"), i);
}
