/* -*- tab-width: 4 -*- */

/* Code that is common to both garbage collectors.
 */

#include "global.h"
#include "elkkernel.h"

int GC_In_Progress;

GCNODE *GC_List;

static GCNODE *Global_GC_Obj;

static FUNCT *Before_GC_Funcs, *After_GC_Funcs;

static ELKObject V_Garbage_Collect_Notifyp;
static ELKObject Sym_Stop_And_Copy_GC, Sym_Generational_GC, Sym_Incremental_GC;

void Visit(ELKObject *p);
void Call_Before_GC (void);
void Call_After_GC (void);
void Visit_GC_List (GCNODE *list, int delta);
void Visit_Wind (WIND *list, unsigned delta);

void Init_Heap (void) {
	Define_Variable (&V_Garbage_Collect_Notifyp, x_("garbage-collect-notify?"),
		Trueob);

	Define_Symbol (&Sym_Stop_And_Copy_GC, x_("stop-and-copy"));
	Define_Symbol (&Sym_Generational_GC, x_("generational"));
	Define_Symbol (&Sym_Incremental_GC, x_("incremental"));
}

void Register_Before_GC (void (*f)(void)) {
	FUNCT *p;

	p = (FUNCT *)Safe_Malloc (sizeof (*p));
	p->func = f;
	p->next = Before_GC_Funcs;
	Before_GC_Funcs = p;
}

void Call_Before_GC (void) {
	FUNCT *p;

	for (p = Before_GC_Funcs; p; p = p->next)
		p->func();
}

void Register_After_GC (void (*f)(void)) {
	FUNCT *p;

	p = (FUNCT *)Safe_Malloc (sizeof (*p));
	p->func = f;
	p->next = After_GC_Funcs;
	After_GC_Funcs = p;
}

void Call_After_GC (void) {
	FUNCT *p;

	for (p = After_GC_Funcs; p; p = p->next)
		p->func();
}

void Visit_GC_List (GCNODE *list, int delta) {
	register GCNODE *gp, *p;
	register int n;
	register ELKObject *vec;

	for (gp = list; gp; gp = p->next) {
		p = (GCNODE *)NORM(gp);
		if (p->gclen <= 0) {
			Visit ((ELKObject *)NORM(p->gcobj));
		} else {
			vec = (ELKObject *)NORM(p->gcobj);
			for (n = 0; n < p->gclen-1; n++)
				Visit (&vec[n]);
		}
	}
}

void Visit_Wind (WIND *list, unsigned int delta) {
	register WIND *wp, *p;

	for (wp = list; wp; wp = p->next) {
		p = (WIND *)NORM(wp);
		Visit (&p->inout);
	}
}

void Func_Global_GC_Link (ELKObject *x) {
	GCNODE *p;

	p = (GCNODE *)Safe_Malloc (sizeof (*p));
	p->gclen = 0;
	p->gcobj = x;
	p->next = Global_GC_Obj;
	Global_GC_Obj = p;
}

#define GC_STRAT_SAC   1
#define GC_STRAT_GEN   2

#define GC_FLAGS_INCR  1

ELKObject Internal_GC_Status(int, int);

ELKObject P_Garbage_Collect_Status (int argc, ELKObject *argv) {
	int strat = 0, flags = 0;

	if (argc > 0) {
		Check_Type (argv[0], T_Symbol);
		if (EQ (argv[0], Sym_Stop_And_Copy_GC))
			strat = GC_STRAT_SAC;
		else if (EQ (argv[0], Sym_Generational_GC))
			strat = GC_STRAT_GEN;
		else Primitive_Error (x_("unknown GC strategy: ~s"), argv[0]);
		if (argc == 2) {
			Check_Type (argv[1], T_Symbol);
			if (EQ (argv[1], Sym_Incremental_GC))
				flags = GC_FLAGS_INCR;
			else Primitive_Error (x_("unknown GC strategy: ~s"), argv[1]);
		}
	}
	return Internal_GC_Status (strat, flags);
}

#ifdef GENERATIONAL_GC
#  include "elkheap-gen.c"
#else
#  include "elkheap-sc.c"
#endif
