/* -*- tab-width: 4 -*- */

#include "elkkernel.h"

ELKObject Cxr (ELKObject x, CHAR *patpat, int len);
ELKObject General_Member (ELKObject key, ELKObject list, int comp);

ELKObject Const_Cons (ELKObject car, ELKObject cdr) {
	ELKObject ret;

	ret = P_Cons (car, cdr);
	SETCONST(ret);
	return ret;
}

ELKObject P_Cons (ELKObject car, ELKObject cdr) {
	ELKObject cell;
	GC_Node2;

#ifdef GENERATIONAL_GC
	GC_Link2 (car, cdr);
	cell = Alloc_Object (sizeof (struct S_Pair), T_Pair, 0);
	GC_Unlink;
#else
	/* This is an optimization (it duplicates parts of Alloc_Object()):
	 */

	register CHAR *p;

	p = Hp;
	ALIGN(p);
	if (p + sizeof (struct S_Pair) <= Heap_End && !GC_Debug) {
		Hp = p + sizeof (struct S_Pair);
		SET(cell, T_Pair, (struct S_Pair *)p);
	} else {
		GC_Link2 (car, cdr);
		cell = Alloc_Object (sizeof (struct S_Pair), T_Pair, 0);
		GC_Unlink;
	}
#endif
	Car (cell) = car;
	Cdr (cell) = cdr;
	return cell;
}

ELKObject P_Car (ELKObject x) {
	Check_Type (x, T_Pair);
	return Car (x);
}

ELKObject P_Cdr (ELKObject x) {
	Check_Type (x, T_Pair);
	return Cdr (x);
}

ELKObject Cxr (ELKObject x, register CHAR *pat, register int len) {
	ELKObject ret;

	for (ret = x, pat += len; len > 0; len--)
		switch (*--pat) {
		case 'a': ret = P_Car (ret); break;
		case 'd': ret = P_Cdr (ret); break;
		default: Primitive_Error (x_("invalid pattern"));
		}
	return ret;
}

ELKObject P_Cddr   (ELKObject x) { return Cxr (x,  x_("dd"), 2); }
ELKObject P_Cdar   (ELKObject x) { return Cxr (x,  x_("da"), 2); }
ELKObject P_Cadr   (ELKObject x) { return Cxr (x,  x_("ad"), 2); }
ELKObject P_Caar   (ELKObject x) { return Cxr (x,  x_("aa"), 2); }

ELKObject P_Cdddr  (ELKObject x) { return Cxr (x, x_("ddd"), 3); }
ELKObject P_Cddar  (ELKObject x) { return Cxr (x, x_("dda"), 3); }
ELKObject P_Cdadr  (ELKObject x) { return Cxr (x, x_("dad"), 3); }
ELKObject P_Cdaar  (ELKObject x) { return Cxr (x, x_("daa"), 3); }
ELKObject P_Caddr  (ELKObject x) { return Cxr (x, x_("add"), 3); }
ELKObject P_Cadar  (ELKObject x) { return Cxr (x, x_("ada"), 3); }
ELKObject P_Caadr  (ELKObject x) { return Cxr (x, x_("aad"), 3); }
ELKObject P_Caaar  (ELKObject x) { return Cxr (x, x_("aaa"), 3); }

ELKObject P_Caaaar (ELKObject x) { return Cxr (x, x_("aaaa"), 4); }
ELKObject P_Caaadr (ELKObject x) { return Cxr (x, x_("aaad"), 4); }
ELKObject P_Caadar (ELKObject x) { return Cxr (x, x_("aada"), 4); }
ELKObject P_Caaddr (ELKObject x) { return Cxr (x, x_("aadd"), 4); }
ELKObject P_Cadaar (ELKObject x) { return Cxr (x, x_("adaa"), 4); }
ELKObject P_Cadadr (ELKObject x) { return Cxr (x, x_("adad"), 4); }
ELKObject P_Caddar (ELKObject x) { return Cxr (x, x_("adda"), 4); }
ELKObject P_Cadddr (ELKObject x) { return Cxr (x, x_("addd"), 4); }
ELKObject P_Cdaaar (ELKObject x) { return Cxr (x, x_("daaa"), 4); }
ELKObject P_Cdaadr (ELKObject x) { return Cxr (x, x_("daad"), 4); }
ELKObject P_Cdadar (ELKObject x) { return Cxr (x, x_("dada"), 4); }
ELKObject P_Cdaddr (ELKObject x) { return Cxr (x, x_("dadd"), 4); }
ELKObject P_Cddaar (ELKObject x) { return Cxr (x, x_("ddaa"), 4); }
ELKObject P_Cddadr (ELKObject x) { return Cxr (x, x_("ddad"), 4); }
ELKObject P_Cdddar (ELKObject x) { return Cxr (x, x_("ddda"), 4); }
ELKObject P_Cddddr (ELKObject x) { return Cxr (x, x_("dddd"), 4); }

ELKObject P_Cxr (ELKObject x, ELKObject pat) {
	Check_List (x);
	if (TYPE(pat) == T_Symbol)
		pat = SYMBOL(pat)->name;
	else if (TYPE(pat) != T_String)
		Wrong_Type_Combination (pat, x_("string or symbol"));
	return Cxr (x, STRING(pat)->data, STRING(pat)->size);
}

ELKObject P_Nullp (ELKObject x) {
	return Nullp (x) ? Trueob : Falseob;
}

ELKObject P_Pairp (ELKObject x) {
	return TYPE(x) == T_Pair ? Trueob : Falseob;
}

ELKObject P_Listp (ELKObject x) {
	ELKObject s;
	register int f;

	for (s = x, f = 0; !Nullp (x); f ^= 1) {
		if (TYPE(x) != T_Pair)
			return Falseob;
		x = Cdr (x);
		if (EQ(x, s))
			return Falseob;
		if (f) s = Cdr (s);
	}
	return Trueob;
}

ELKObject P_Set_Car (ELKObject x, ELKObject newv) {
	Check_Type (x, T_Pair);
	Check_Mutable (x);
	Car (x) = newv;
	return newv;
}

ELKObject P_Set_Cdr (ELKObject x, ELKObject newv) {
	Check_Type (x, T_Pair);
	Check_Mutable (x);
	Cdr (x) = newv;
	return newv;
}

ELKObject General_Member (ELKObject key, ELKObject list, register int comp) {
	register int r;

	for ( ; !Nullp (list); list = Cdr (list)) {
		Check_List (list);
		if (comp == 0)
			r = EQ(Car (list), key);
		else if (comp == 1)
			r = Eqv (Car (list), key);
		else
			r = Equal (Car (list), key);
		if (r) return list;
	}
	return Falseob;
}

ELKObject P_Memq (ELKObject key, ELKObject list) {
	return General_Member (key, list, 0);
}

ELKObject P_Memv (ELKObject key, ELKObject list) {
	return General_Member (key, list, 1);
}

ELKObject P_Member (ELKObject key, ELKObject list) {
	return General_Member (key, list, 2);
}

ELKObject General_Assoc (ELKObject key, ELKObject alist, register int comp) {
	ELKObject elem;
	register int r;

	for ( ; !Nullp (alist); alist = Cdr (alist)) {
		Check_List (alist);
		elem = Car (alist);
		if (TYPE(elem) != T_Pair)
			continue;
		if (comp == 0)
			r = EQ(Car (elem), key);
		else if (comp == 1)
			r = Eqv (Car (elem), key);
		else
			r = Equal (Car (elem), key);
		if (r) return elem;
	}
	return Falseob;
}

ELKObject P_Assq (ELKObject key, ELKObject alist) {
	return General_Assoc (key, alist, 0);
}

ELKObject P_Assv (ELKObject key, ELKObject alist) {
	return General_Assoc (key, alist, 1);
}

ELKObject P_Assoc (ELKObject key, ELKObject alist) {
	return General_Assoc (key, alist, 2);
}

int Fast_Length (ELKObject list) {
	ELKObject tail;
	register int i;

	for (i = 0, tail = list; TYPE(tail) == T_Pair; tail = Cdr (tail), i++)
		;
	return i;
}

ELKObject P_Length (ELKObject list) {
	ELKObject tail;
	register int i;

	for (i = 0, tail = list; !Nullp (tail); tail = Cdr (tail), i++)
		Check_List (tail);
	return Make_Integer (i);
}

ELKObject P_Make_List (ELKObject n, ELKObject init) {
	register int len;
	ELKObject list;
	GC_Node;

	if ((len = Get_Exact_Integer (n)) < 0)
		Range_Error (n);
	list = Null;
	GC_Link (init);
	while (len-- > 0)
		list = Cons (init, list);
	GC_Unlink;
	return list;
}

ELKObject P_List (int argc, ELKObject *argv) {
	ELKObject list, tail, cell;
	GC_Node2;

	GC_Link2 (list, tail);
	for (list = tail = Null; argc-- > 0; tail = cell) {
		cell = Cons (*argv++, Null);
		if (Nullp (list))
			list = cell;
		else
			(void)P_Set_Cdr (tail, cell);
	}
	GC_Unlink;
	return list;
}

ELKObject P_Last_Pair (ELKObject x) {
	Check_Type (x, T_Pair);
	for ( ; TYPE(Cdr (x)) == T_Pair; x = Cdr (x)) ;
	return x;
}

ELKObject P_Append (int argc, ELKObject *argv) {
	ELKObject list, last, tail, cell;
	register int i;
	GC_Node3;

	list = last = Null;
	GC_Link3 (list, last, tail);
	for (i = 0; i < argc-1; i++) {
		for (tail = argv[i]; !Nullp (tail); tail = Cdr (tail)) {
			Check_List (tail);
			cell = Cons (Car (tail), Null);
			if (Nullp (list))
				list = cell;
			else
				(void)P_Set_Cdr (last, cell);
			last = cell;
		}
	}
	if (argc)
		if (Nullp (list))
			list = argv[i];
		else
			(void)P_Set_Cdr (last, argv[i]);
	GC_Unlink;
	return list;
}

ELKObject P_Append_Set (int argc, ELKObject *argv) {
	register int i, j;

	for (i = j = 0; i < argc; i++)
		if (!Nullp (argv[i]))
			argv[j++] = argv[i];
	if (j == 0)
		return Null;
	for (i = 0; i < j-1; i++)
		(void)P_Set_Cdr (P_Last_Pair (argv[i]), argv[i+1]);
	return *argv;
}

ELKObject P_Reverse (ELKObject x) {
	ELKObject ret;
	GC_Node;

	GC_Link (x);
	for (ret = Null; !Nullp (x); x = Cdr (x)) {
		Check_List (x);
		ret = Cons (Car (x), ret);
	}
	GC_Unlink;
	return ret;
}

ELKObject P_Reverse_Set (ELKObject x) {
	ELKObject prev, tail;

	for (prev = Null; !Nullp (x); prev = x, x = tail) {
		Check_List (x);
		tail = Cdr (x);
		(void)P_Set_Cdr (x, prev);
	}
	return prev;
}

ELKObject P_List_Tail (ELKObject x, ELKObject num) {
	register int n;

	for (n = Get_Exact_Integer (num); n > 0 && !Nullp (x); n--, x = P_Cdr (x))
		;
	return x;
}

ELKObject P_List_Ref (ELKObject x, ELKObject num) {
	return P_Car (P_List_Tail (x, num));
}

ELKObject Copy_List (ELKObject x) {
	ELKObject car, cdr;
	GC_Node3;

	if (TYPE(x) == T_Pair) {
		if (Stack_Size () > Max_Stack)
			Uncatchable_Error (x_("Out of stack space"));
		car = cdr = Null;
		GC_Link3 (x, car, cdr);
		car = Copy_List (Car (x));
		cdr = Copy_List (Cdr (x));
		x = Cons (car, cdr);
		GC_Unlink;
	}
	return x;
}
