/* -*- tab-width: 4 -*- */

/* Read and manage symbol tables from object modules.
 */

#include "elkkernel.h"

#if defined(CAN_LOAD_OBJ) || defined (INIT_OBJECTS)

#ifdef MACH_O
#  include "stab-macho.c"
#else
#ifdef ELF
#  include "stab-elf.c"
#else
#if defined(COFF) || defined(XCOFF)
#  include "stab-coff.c"
#else
#ifdef ECOFF
#  include "stab-ecoff.c"
#else
#ifdef CONVEX_AOUT
#  include "stab-convex.c"
#else
#if defined(hp9000s300) || defined(__hp9000s300) || defined(__hp9000s300__)
#  include "stab-hp9k300.c"
#else
#if defined(hp9000s800) || defined(__hp9000s800) || defined(__hp9000s800__)
#  include "stab-hp9k800.c"
#else
#  include "stab-bsd.c"
#endif
#endif
#endif
#endif
#endif
#endif
#endif

static SYMPREFIX Ignore_Prefixes[] =  {
	/* Currently none */
	0, 0
};
static SYMPREFIX Init_Prefixes[] =  {
	INIT_PREFIX,	PR_EXTENSION,
	x_("_GLOBAL_.I."),  PR_CONSTRUCTOR,	/* SVR4.2/g++ */
	x_("__sti__"),	  PR_CONSTRUCTOR,
	x_("_STI"),		 PR_CONSTRUCTOR,
	x_("_GLOBAL_$I$"),  PR_CONSTRUCTOR,
	0, 0
};
static SYMPREFIX Finit_Prefixes[] = {
	FINIT_PREFIX,   PR_EXTENSION,
	x_("_GLOBAL_.D."),  PR_CONSTRUCTOR,
	x_("__std__"),	  PR_CONSTRUCTOR,
	x_("_STD"),		 PR_CONSTRUCTOR,
	x_("_GLOBAL_$D$"),  PR_CONSTRUCTOR,
	0, 0
};

static FUNCT *Finalizers;

static void Call (unsigned long l) {
#ifdef XCOFF
	unsigned long vec[3];
	extern main();

	bcopy ((CHAR *)main, (CHAR *)vec, sizeof vec);
	vec[0] = (l & ~0xF0000000) + (vec[0] & 0xF0000000);
	((void (*)())vec)();
#else
	((void (*)())l)();
#endif
}

Call_Initializers (SYMTAB *tab, CHAR *addr, int which) {
	SYM *sp;
	CHAR *p;
	SYMPREFIX *pp;
	FUNCT *fp, **fpp;

	/* Set pointer to end of list of finalizers; extension finalization
	 * functions and C++ static destructors will be appended to this list:
	 */
	for (fpp = &Finalizers; *fpp; fpp = &(*fpp)->next)
		;

	for (sp = tab->first; sp; sp = sp->next) {
		if ((CHAR *)sp->value < addr)
			continue;
		p = sp->name;
#ifdef SYMS_BEGIN_WITH
		if (*p == SYMS_BEGIN_WITH)
			p++;
		else
			continue;
#endif
		for (pp = Ignore_Prefixes; pp->name; pp++)
			if (estrncmp (p, pp->name, estrlen (pp->name)) == 0)
				goto next;
		for (pp = Init_Prefixes; pp->name; pp++) {
			if (pp->type == which
					&& estrncmp (p, pp->name, estrlen (pp->name)) == 0) {
				if (Verb_Init)
					ttyputmsg(x_("[calling %s]"), p);
				Call (sp->value);
			}
		}
		/* Append to list of finalizers (to be invoked on exit):
		 */
		for (pp = Finit_Prefixes; pp->name; pp++) {
			if (pp->type == which
					&& estrncmp (p, pp->name, estrlen (pp->name)) == 0) {
				fp = (FUNCT *)Safe_Malloc (sizeof (FUNCT));
				fp->func = (void (*)())sp->value;
				fp->name = Safe_Malloc (estrlen (p) + 1);
				estrcpy (fp->name, p);
				fp->next = 0;
				*fpp = fp;
				fpp = &fp->next;
			}
		}
next: ;
	}
}

/* Call the finialization functions and C++ static destructors.  Make sure
 * that calling exit() from a function doesn't cause endless recursion.
 */
Call_Finalizers (void) {
	while (Finalizers) {
		FUNCT *fp = Finalizers;
		Finalizers = fp->next;
		if (Verb_Init)
			ttyputmsg(x_("[calling %s]"), fp->name);
		Call ((unsigned long)fp->func);
	}
}

Free_Symbols (SYMTAB *tab) {
	register SYM *sp, *nextp;

	for (sp = tab->first; sp; sp = nextp) {
		nextp = sp->next;
#if defined(COFF) || defined(ECOFF)
		free (sp->name);
#endif
		free ((CHAR *)sp);
	}
	if (tab->strings)
		free (tab->strings);
	free ((CHAR *)tab);
}
#endif /* CAN_LOAD_OBJ || INIT_OBJECTS */
