/* -*- tab-width: 4 -*- */

#include <ctype.h>

#include "elkkernel.h"

CHAR Char_Map[256];
ELKObject General_Substringp (ELKObject s1, ELKObject s2, int ci);
int General_Strcmp (ELKObject s1, ELKObject s2, int ci);

void Init_String (void) {
	register int i;

	for (i = 0; i < 256; i++)
		Char_Map[i] = i;
	for (i = 'A'; i <= 'Z'; i++)
		Char_Map[i] = tolower (i);
}

ELKObject General_Make_String (const CHAR *s, int len, int konst) {
	ELKObject str;

	str = Alloc_Object (len + sizeof (struct S_String) - 1, T_String, konst);
	STRING(str)->tag = Null;
	STRING(str)->size = len;
	if (s)
		bcopy (s, STRING(str)->data, len);
	return str;
}

ELKObject Make_String (const CHAR *s, int len) {
	return General_Make_String (s, len, 0);
}

ELKObject Make_Const_String (const CHAR *s, int len) {
	return General_Make_String (s, len, 1);
}

ELKObject P_Stringp (ELKObject s) {
	return TYPE(s) == T_String ? Trueob : Falseob;
}

ELKObject P_Make_String (int argc, ELKObject *argv) {
	register int len, c = ' ';
	ELKObject str;
	register CHAR *p;

	if ((len = Get_Exact_Integer (argv[0])) < 0)
		Range_Error (argv[0]);
	if (argc == 2) {
		Check_Type (argv[1], T_Character);
		c = ELKCHAR(argv[1]);
	}
	str = Make_String ((CHAR *)0, len);
	for (p = STRING(str)->data; len; len--) *p++ = c;
	return str;
}

ELKObject P_String (int argc, ELKObject *argv) {
	ELKObject str;
	register int i;

	str = Make_String ((CHAR *)0, argc);
	for (i = 0; i < argc; i++) {
		Check_Type (argv[i], T_Character);
		STRING(str)->data[i] = ELKCHAR(argv[i]);
	}
	return str;
}

ELKObject P_String_To_Number (int argc, ELKObject *argv) {
	ELKObject ret;
	CHAR *b;
	register struct S_String *p;
	int radix = 10;
	Alloca_Begin;

	Check_Type (argv[0], T_String);
	if (argc == 2) {
		radix = Get_Exact_Integer (argv[1]);
		switch (radix) {
		case 2: case 8: case 10: case 16:
			break;
		default:
			Primitive_Error (x_("invalid radix: ~s"), argv[1]);
		}
	}
	p = STRING(argv[0]);
	Alloca (b, CHAR*, p->size+1);
	bcopy (p->data, b, p->size);
	b[p->size] = '\0';
	ret = Parse_Number (Null, b, radix);
	Alloca_End;
	return Nullp (ret) ? Falseob : ret;
}

ELKObject P_String_Length (ELKObject s) {
	Check_Type (s, T_String);
	return Make_Integer (STRING(s)->size);
}

ELKObject P_String_Ref (ELKObject s, ELKObject n) {
	Check_Type (s, T_String);
	return Make_Char (STRING(s)->data[Get_Index (n, s)]);
}

ELKObject P_String_Set (ELKObject s, ELKObject n, ELKObject newv) {
	register int i, old;

	Check_Type (s, T_String);
	Check_Mutable (s);
	Check_Type (newv, T_Character);
	old = STRING(s)->data[i = Get_Index (n, s)];
	STRING(s)->data[i] = ELKCHAR(newv);
	return Make_Char (old);
}

ELKObject P_Substring (ELKObject s, ELKObject a, ELKObject b) {
	register int i, j;

	Check_Type (s, T_String);
	if ((i = Get_Exact_Integer (a)) < 0 || i > STRING(s)->size)
		Range_Error (a);
	if ((j = Get_Exact_Integer (b)) < 0 || j > STRING(s)->size)
		Range_Error (b);
	if (i > j)
		Primitive_Error (x_("`end' less than `start'"));
	return Make_String (&STRING(s)->data[i], j-i);
}

ELKObject P_String_Copy (ELKObject s) {
	Check_Type (s, T_String);
	return Make_String (STRING(s)->data, STRING(s)->size);
}

ELKObject P_String_Append (int argc, ELKObject *argv) {
	register int i, len;
	ELKObject s, str;

	for (len = i = 0; i < argc; i++) {
		Check_Type (argv[i], T_String);
		len += STRING(argv[i])->size;
	}
	str = Make_String ((CHAR *)0, len);
	for (len = i = 0; i < argc; i++) {
		s = argv[i];
		bcopy (STRING(s)->data, &STRING(str)->data[len], STRING(s)->size);
		len += STRING(s)->size;
	}
	return str;
}

ELKObject P_List_To_String (ELKObject list) {
	ELKObject str, len;
	register int i;
	GC_Node;

	GC_Link (list);
	len = P_Length (list);
	str = Make_String ((CHAR *)0, FIXNUM(len));
	for (i = 0; i < FIXNUM(len); i++, list = Cdr (list)) {
		Check_Type (Car (list), T_Character);
		STRING(str)->data[i] = ELKCHAR(Car (list));
	}
	GC_Unlink;
	return str;
}

ELKObject P_String_To_List (ELKObject s) {
	register int i;
	ELKObject list, tail, cell;
	GC_Node3;

	Check_Type (s, T_String);
	list = tail = Null;
	GC_Link3 (s, list, tail);
	for (i = 0; i < STRING(s)->size; i++, tail = cell) {
		cell = Cons (Make_Char (STRING(s)->data[i]), Null);
		if (Nullp (list))
			list = cell;
		else
			(void)P_Set_Cdr (tail, cell);
	}
	GC_Unlink;
	return list;
}

ELKObject P_Substring_Fill (ELKObject s, ELKObject a, ELKObject b, ELKObject c) {
	register int i, j;

	Check_Type (s, T_String);
	Check_Mutable (s);
	Check_Type (c, T_Character);
	i = Get_Index (a, s);
	if ((j = Get_Exact_Integer (b)) < 0 || j > STRING(s)->size)
		Range_Error (b);
	if (i > j)
		Primitive_Error (x_("`end' less than `start'"));
	while (i < j)
		STRING(s)->data[i++] = ELKCHAR(c);
	return s;
}

ELKObject P_String_Fill (ELKObject s, ELKObject c) {
	ELKObject ret;
	GC_Node2;

	Check_Type (s, T_String);
	Check_Mutable (s);
	GC_Link2 (s, c);
	ret = P_Substring_Fill (s, Make_Integer (0),
		Make_Integer (STRING(s)->size), c);
	GC_Unlink;
	return ret;
}

ELKObject General_Substringp (ELKObject s1, ELKObject s2, register int ci) {
	register int n, l1, l2;
	register CHAR *p1, *p2, *p3, *map;

	Check_Type (s1, T_String);
	Check_Type (s2, T_String);
	l1 = STRING(s1)->size;
	l2 = STRING(s2)->size;
	map = Char_Map;
	for (p2 = STRING(s2)->data; l2 >= l1; p2++, l2--) {
		for (p1 = STRING(s1)->data, p3 = p2, n = l1; n; n--, p1++, p3++) {
			if (ci) {
				if (map[*p1] != map[*p3]) goto fail;
			} else
				if (*p1 != *p3) goto fail;
		}
		return Make_Integer (STRING(s2)->size - l2);
fail:   ;
	}
	return Falseob;
}

ELKObject P_Substringp (ELKObject s1, ELKObject s2) {
	return General_Substringp (s1, s2, 0);
}

ELKObject P_CI_Substringp (ELKObject s1, ELKObject s2) {
	return General_Substringp (s1, s2, 1);
}

int General_Strcmp (ELKObject s1, ELKObject s2, register int ci) {
	register int n, l1, l2;
	register CHAR *p1, *p2, *map;

	Check_Type (s1, T_String);
	Check_Type (s2, T_String);
	l1 = STRING(s1)->size; l2 = STRING(s2)->size;
	n = l1 > l2 ? l2 : l1;
	p1 = STRING(s1)->data; p2 = STRING(s2)->data;
	for (map = Char_Map; --n >= 0; p1++, p2++) {
		if (ci) {
			if (map[*p1] != map[*p2]) break;
		} else
			if (*p1 != *p2) break;
	}
	if (n < 0)
		return l1 - l2;
	return ci ? map[*p1] - map[*p2] : *p1 - *p2;
}

ELKObject P_String_Eq (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 0) ? Falseob : Trueob;
}

ELKObject P_String_Less (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 0) < 0 ? Trueob : Falseob;
}

ELKObject P_String_Greater (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 0) > 0 ? Trueob : Falseob;
}

ELKObject P_String_Eq_Less (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 0) <= 0 ? Trueob : Falseob;
}

ELKObject P_String_Eq_Greater (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 0) >= 0 ? Trueob : Falseob;
}

ELKObject P_String_CI_Eq (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 1) ? Falseob : Trueob;
}

ELKObject P_String_CI_Less (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 1) < 0 ? Trueob : Falseob;
}

ELKObject P_String_CI_Greater (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 1) > 0 ? Trueob : Falseob;
}

ELKObject P_String_CI_Eq_Less (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 1) <= 0 ? Trueob : Falseob;
}

ELKObject P_String_CI_Eq_Greater (ELKObject s1, ELKObject s2) {
	return General_Strcmp (s1, s2, 1) >= 0 ? Trueob : Falseob;
}
