/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

/* 
 * Auxiliary net macros.
 */
 
#ifndef	NCHAN
#  include "irsimnet.h"
#endif

#define SWAP( TYPE, A, B )		\
  {					\
    register TYPE TMP;			\
					\
    TMP = (A);				\
    (A) = (B);				\
    (B) = TMP;				\
  }

/*
 * Swap the 2 nodes
 */
#define	SWAP_NODES(N1, N2)		SWAP(nptr, N1, N2)


/*
 * if FLAG is not set in in NODE->nflags, Link NODE to the head of LIST
 * using the temporary entry (n.next) in the node structure.  This is
 * used during net read-in/change to build lists of affected nodes.
 */
#define	LINK_TO_LIST(NODE, LIST, FLAG)	\
  {						\
    if (((NODE)->nflags & (FLAG)) == 0)	\
      {						\
	(NODE)->nflags |= (FLAG);		\
	(NODE)->n.next = (LIST);		\
	LIST = (NODE);				\
      }						\
  }						\



/*
 * Allocate a new "Tlist" pointer.
 */
#define	NEW_LINK(LP)						\
  {								\
    if ((LP = irsim_freeLinks) == NULL)				\
	LP = (lptr)irsim_MallocList(sizeof(struct Tlist), 1);	\
    irsim_freeLinks = (LP)->next;					\
  }


/*
 * Return "Tlist" pointer LP to free pool.
 */
#define	FREE_LINK(LP)			\
  {					\
    (LP)->next = irsim_freeLinks;		\
    irsim_freeLinks = (LP);			\
  }


/*
 * Allocate a new Transistor.
 */
#ifndef USER_SUBCKT
#define	NEW_TRANS(T)						\
  {								\
    if ((T = irsim_freeTrans) == NULL)				\
	T = (tptr) irsim_MallocList(sizeof(struct Trans), 1);	\
    irsim_freeTrans = (tptr) (T)->gate;				\
  }
#else 
#define	NEW_TRANS(T)						\
  {								\
    if ((T = irsim_freeTrans) == NULL)				\
	T = (tptr) irsim_MallocList(sizeof(struct Trans), 1);	\
    T->subptr = NULL ;                                          \
    irsim_freeTrans = (tptr) (T)->gate;				\
  }
#define NEW_SUBCKT(S) \
  { \
   S = (Subckt *) emalloc(sizeof (Subckt), sim_tool->cluster);\
   bzero(S , sizeof(Subckt));\
  }
#endif


/*
 * Return transistor record T to free pool.
 */
#define	FREE_TRANS(T)			\
  {					\
    (T)->gate = (nptr) irsim_freeTrans;	\
    irsim_freeTrans = (T);			\
  }					\


/*
 * Add transistor T to the list of transistors connected to that list.
 * The transistor is added at the head of the list.
 */
#define CONNECT(LIST, T)	\
  {				\
    register lptr  newl;	\
				\
    NEW_LINK(newl);		\
    newl->xtor = (T);		\
    newl->next = (LIST);	\
    LIST = newl;		\
  }


/*
 * Transistors that have their drain/source shorted are NOT connected
 * to the network, they are instead linked as a doubly linked list
 * using the scache/dcache fields.
 */
#define LINK_TCAP(T)			\
  {					\
    (T)->dcache.t = irsim_tcap;		\
    (T)->scache.t = irsim_tcap->scache.t;	\
    irsim_tcap->scache.t->dcache.t = (T);	\
    irsim_tcap->scache.t = (T);		\
    irsim_tcap->x.pos ++;			\
  }


#define	UNLINK_TCAP(T)			\
  {						\
    (T)->dcache.t->scache.t = (T)->scache.t;	\
    (T)->scache.t->dcache.t = (T)->dcache.t;	\
    (T)->ttype &= ~TCAP;			\
    irsim_tcap->x.pos --;				\
  }


/*
 * Replace the first ocurrence of transistor OLD by NEW on LIST.
 */
#define	REPLACE(LIST, OLD, NEW)			\
  {							\
    register lptr  lp;					\
							\
    for(lp = (LIST); lp != NULL; lp = lp->next)	\
      {							\
	if (lp->xtor == OLD)				\
	  {						\
	    lp->xtor = NEW;				\
	    break;					\
	  }						\
      }							\
  }							\


/*
 * Remove the entry for transistor T from LIST, and return it to the
 * free pool.
 */
#define	DISCONNECT(LIST, T)			\
  {						\
    register lptr  li, *lip;			\
						\
    lip = &(LIST);				\
    while((li = *lip) != NULL)		\
      {						\
	if(li->xtor == (T))			\
	  {					\
	    *lip = li->next;			\
	    FREE_LINK(li);			\
	    break;				\
	  }					\
	lip = &(li->next);			\
      }						\
  }						

